import json
from urllib.parse import urlencode
from urllib.request import ProxyHandler, Request, build_opener

from django.conf import settings

from drf_recaptcha.constants import DEFAULT_RECAPTCHA_DOMAIN


class RecaptchaResponse:
    def __init__(self, is_valid, error_codes=None, extra_data=None):
        self.is_valid = is_valid
        self.error_codes = error_codes or []
        self.extra_data = extra_data or {}


def recaptcha_request(params):
    request_object = Request(
        url="https://%s/recaptcha/api/siteverify"
        % getattr(settings, "DRF_RECAPTCHA_DOMAIN", DEFAULT_RECAPTCHA_DOMAIN),
        data=params,
        headers={
            "Content-type": "application/x-www-form-urlencoded",
            "User-agent": "DRF reCAPTCHA",
        },
    )

    # Add proxy values to opener if needed.
    opener_args = []
    proxies = getattr(settings, "DRF_RECAPTCHA_PROXY", {})
    if proxies:
        opener_args = [ProxyHandler(proxies)]
    opener = build_opener(*opener_args)

    # Get response from POST to Google endpoint.
    return opener.open(
        request_object,
        timeout=getattr(settings, "DRF_RECAPTCHA_VERIFY_REQUEST_TIMEOUT", 10),
    )


def submit(recaptcha_response, secret_key, remoteip):
    params = urlencode(
        {"secret": secret_key, "response": recaptcha_response, "remoteip": remoteip}
    )

    params = params.encode("utf-8")

    response = recaptcha_request(params)
    data = json.loads(response.read().decode("utf-8"))
    response.close()
    return RecaptchaResponse(
        is_valid=data.pop("success"),
        error_codes=data.pop("error-codes", None),
        extra_data=data,
    )



# Test keys https://developers.google.com/recaptcha/docs/faq
#
# For reCAPTCHA v3, create a separate key for testing environments.
# Scores may not be accurate as reCAPTCHA v3 relies on seeing real traffic.
#
# For reCAPTCHA v2, use the following test keys.
# You will always get No CAPTCHA and all verification requests will pass.

TEST_V2_SECRET_KEY = "6LeIxAcTAAAAAGG-vFI1TnRWxMZNFuojJ4WifJWe"

DEFAULT_RECAPTCHA_DOMAIN = "www.google.com"

# https://developers.google.com/recaptcha/docs/v3
#
# reCAPTCHA v3 returns a score (1.0 is very likely a good interaction, 0.0 is very likely a bot).
# Based on the score, you can take variable action in the context of your site.

DEFAULT_V3_SCORE = 0.5



from django.apps import AppConfig


class DRFreCaptchaConfig(AppConfig):
    name = "drf_recaptcha"
    verbose_name = "Django REST framework reCAPTCHA"

    def ready(self):
        # Add System checks
        from .checks import recaptcha_system_check  # NOQA



from django.conf import settings
from rest_framework.serializers import CharField

from drf_recaptcha.constants import DEFAULT_V3_SCORE
from drf_recaptcha.validators import ReCaptchaV2Validator, ReCaptchaV3Validator


class ReCaptchaV2Field(CharField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.write_only = True

        validator = ReCaptchaV2Validator(secret_key=settings.DRF_RECAPTCHA_SECRET_KEY)
        self.validators.append(validator)


class ReCaptchaV3Field(CharField):
    def __init__(self, action: str, required_score: float = None, **kwargs):
        super().__init__(**kwargs)

        self.write_only = True

        scores_from_settings = getattr(settings, "DRF_RECAPTCHA_ACTION_V3_SCORES", {})

        required_score = (
            required_score
            or getattr(scores_from_settings, action, None)
            or getattr(settings, "DRF_RECAPTCHA_DEFAULT_V3_SCORE", DEFAULT_V3_SCORE)
        )

        validator = ReCaptchaV3Validator(
            action=action,
            required_score=required_score,
            secret_key=settings.DRF_RECAPTCHA_SECRET_KEY,
        )
        self.validators.append(validator)



from django.conf import settings
from django.core.checks import Tags, Warning, register
from django.core.exceptions import ImproperlyConfigured

from drf_recaptcha.constants import TEST_V2_SECRET_KEY


@register(Tags.security)
def recaptcha_system_check(app_configs, **kwargs):
    errors = []

    is_testing = getattr(settings, "DRF_RECAPTCHA_TESTING", False)
    if is_testing:
        return errors

    secret_key = getattr(settings, "DRF_RECAPTCHA_SECRET_KEY", None)
    if not secret_key:
        raise ImproperlyConfigured("settings.DRF_RECAPTCHA_SECRET_KEY must be set.")

    if secret_key == TEST_V2_SECRET_KEY:
        errors.append(
            Warning(
                "Google test key for reCAPTCHA v2 is used now.\n"
                "If you use reCAPTCHA v2 - you will always get No CAPTCHA and all"
                " verification requests will pass.\n"
                "If you use reCAPTCHA v3 - all verification requests will fail.",
                hint="Update settings.DRF_RECAPTCHA_SECRET_KEY",
                id="drf_recaptcha.recaptcha_test_key_error",
            )
        )
    return errors



from setuptools import find_packages, setup


def read(f):
    return open(f, "r", encoding="utf-8").read()


setup(
    name="drf-recaptcha",
    version="1.3.0",
    description="Django rest framework recaptcha field serializer.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Lev Lybin",
    author_email="lev.lybin@gmail.com",
    license="MIT",
    url="https://github.com/llybin/drf-recaptcha",
    packages=find_packages(exclude=["tests*", "manage.py"]),
    install_requires=[
        "django>2.0,<4.0",
        "djangorestframework>=3.10,<4.0",
        "django-ipware>=2.1.0,<3.0",
    ],
    python_requires=">=3.5",
    include_package_data=True,
    zip_safe=False,
    keywords=[
        "django",
        "drf",
        "django-rest-framework",
        "reCAPTCHA",
        "reCAPTCHA v2",
        "reCAPTCHA v3",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import logging
from urllib.error import HTTPError

from django.conf import settings
from ipware import get_client_ip
from rest_framework.serializers import ValidationError

from drf_recaptcha import client

logger = logging.getLogger(__name__)


class ReCaptchaValidator:
    messages = {
        "captcha_invalid": "Error verifying reCAPTCHA, please try again.",
        "captcha_error": "Error verifying reCAPTCHA, please try again.",
    }
    recaptcha_client_ip = ""
    recaptcha_secret_key = ""

    def is_testing_and_pass(self) -> bool:
        is_testing = getattr(settings, "DRF_RECAPTCHA_TESTING", False)
        if is_testing:
            testing_result = getattr(settings, "DRF_RECAPTCHA_TESTING_PASS", True)
            if not testing_result:
                raise ValidationError(
                    self.messages["captcha_invalid"], code="captcha_invalid"
                )
            else:
                return True

        return False

    def set_context(self, serializer_field):
        request = serializer_field.context.get("request")

        try:
            self.recaptcha_client_ip = get_client_ip(request)
        except AttributeError:
            logger.exception(
                "Couldn't get client ip address. Check your serializer gets context with request."
            )

    def get_response(self, value: str) -> client.RecaptchaResponse:
        try:
            check_captcha = client.submit(
                recaptcha_response=value,
                secret_key=self.recaptcha_secret_key,
                remoteip=self.recaptcha_client_ip,
            )
        except HTTPError:  # Catch timeouts, etc
            logger.exception("Couldn't get response, HTTPError")
            raise ValidationError(self.messages["captcha_error"], code="captcha_error")

        if not check_captcha.is_valid:
            logger.error(
                "ReCAPTCHA validation failed due to: %s", check_captcha.error_codes
            )
            raise ValidationError(
                self.messages["captcha_invalid"], code="captcha_invalid"
            )

        return check_captcha


class ReCaptchaV2Validator(ReCaptchaValidator):
    def __init__(self, secret_key):
        self.recaptcha_secret_key = secret_key

    def __call__(self, value):
        if self.is_testing_and_pass():
            return

        check_captcha = self.get_response(value)

        score = check_captcha.extra_data.get("score", None)
        if score is not None:
            logger.error(
                "The response contains score, reCAPTCHA v2 response doesn't"
                " contains score, probably secret key for reCAPTCHA v3"
            )
            raise ValidationError(self.messages["captcha_error"], code="captcha_error")


class ReCaptchaV3Validator(ReCaptchaValidator):
    def __init__(self, action, required_score, secret_key):
        self.recaptcha_action = action
        self.recaptcha_required_score = required_score
        self.recaptcha_secret_key = secret_key

    def __call__(self, value):
        if self.is_testing_and_pass():
            return

        check_captcha = self.get_response(value)

        score = check_captcha.extra_data.get("score", None)
        if score is None:
            logger.error(
                "The response not contains score, reCAPTCHA v3 response must"
                " contains score, probably secret key for reCAPTCHA v2"
            )
            raise ValidationError(self.messages["captcha_error"], code="captcha_error")

        action = check_captcha.extra_data.get("action", "")

        if self.recaptcha_required_score >= float(score):
            logger.error(
                "ReCAPTCHA validation failed due to score of %s"
                " being lower than the required amount for action '%s'.",
                score,
                action,
            )
            raise ValidationError(
                self.messages["captcha_invalid"], code="captcha_invalid"
            )

        if self.recaptcha_action != action:
            logger.error(
                "ReCAPTCHA validation failed due to value of action '%s'"
                " is not equal with defined '%s'.",
                action,
                self.recaptcha_action,
            )
            raise ValidationError(
                self.messages["captcha_invalid"], code="captcha_invalid"
            )
